"""A package for already-implemented discriminator models.
"""

from nalp.models.discriminators.conv import ConvDiscriminator
from nalp.models.discriminators.embedded_text import EmbeddedTextDiscriminator
from nalp.models.discriminators.linear import LinearDiscriminator
from nalp.models.discriminators.lstm import LSTMDiscriminator
from nalp.models.discriminators.text import TextDiscriminator
