"""A package for already-implemented machine learning architectures, divided into
discriminators and generators models.
"""

from nalp.models.dcgan import DCGAN
from nalp.models.gan import GAN
from nalp.models.gsgan import GSGAN
from nalp.models.maligan import MaliGAN
from nalp.models.relgan import RelGAN
from nalp.models.seqgan import SeqGAN
from nalp.models.wgan import WGAN
