"""A package for already-implemented machine learning layers.
"""

from nalp.models.layers.gumbel_softmax import GumbelSoftmax
from nalp.models.layers.multi_head_attention import MultiHeadAttention
from nalp.models.layers.relational_memory_cell import RelationalMemoryCell
